#!/usr/bin/env python
#
# $Id: in_mpg123.py,v 1.1.1.1 2003/02/11 06:07:41 resolve Exp $
#
# Started 23/9/01 by Damien Elmes <resolve@repose.cx>
# This file is licensed under the GPL.

"""
plugin for muse, to control mpg123, mpg321, and similar clones which
implement the -R remote interface.
"""

import re
import os
import fcntl
FCNTL=fcntl
import atexit
import signal

import muse
import config
import ui
import multiplex
import input_plugin
import mp3meta

from debug import *

config_defaults = {
    "mpg123_executable": "mpg321"
}

config.set_defaults("in_mpg123", config_defaults)

def sigchld_handler(sig, fr):

    res = os.waitpid(-1, os.WNOHANG)

    # check remote_pid, to make sure it wasn't shut down deliberatiley
    if res[0]: #and self.remote_pid:
        debug("mpg123 remote died. something's very wrong.")
        #self.start_remote()

    signal.signal(signal.SIGCHLD, sigchld_handler)

class new(input_plugin.new):
    """provides an input_plugin capable of playing mp3 files"""

    def __init__(self):
        # call out parent's constuctor
        input_plugin.new.__init__(self)

        # plugin requisite stuff
        self.plugin_type = "input"
        self.plugin_desc = "mpg123/321-remote based plugin"
        self.plugin_name = "mpg123"

        self.file_match = re.compile(r"\.mp[23]$")
        self.status_match = re.compile(r"@F (\d+) (\d+) (\d+.\d+) (\d+.\d+)")
        self.finished_match = re.compile(r"finished.$")

        self.skip_destination = 0
        self.mpg123_executable = "mpg123-oss" # config.get("in_mpg123", "mpg123_executable")

        self.expect_stack = []

        self.start_remote()

        self._reset_state()

        atexit.register(self.stop_remote)

        self.remote_ack=0

    def stop_remote(self):
        os.kill(-self.remote_pid, signal.SIGKILL)

        self.remote_pid = None
        self.remote_msg.close()

    def start_remote(self, player=None):
        """ fork a new decoder process, and set up fds for read+write """

        if player==None: player=self.mpg123_executable

        args = [ player, "-R", "1", "-G 20", "-b 512", \
                 "--aggressive" ]

        (child_stdin_out, child_stdin_in) = os.pipe()
        (child_stdout_out, child_stdout_in) = os.pipe()

        pid = os.fork()

        if pid == 0:
            # we're the child
            # set the session ID, so both mpg123 and this process are killed
            # when a stop is issued

            os.setsid()

            # close unused fds
            os.close(child_stdout_out)
            os.close(child_stdin_in)

            # now dup the fd and load mpg123 on top
            os.dup2(child_stdout_in, os.sys.stdout.fileno())
            os.dup2(child_stdout_in, os.sys.stderr.fileno())
            os.dup2(child_stdin_out, os.sys.stdin.fileno())

            os.execvp(player, args)

        else:
            # we're the parent
            self.remote_from = child_stdout_out
            self.remote_to = child_stdin_in
            self.remote_msg = os.fdopen(child_stdout_out, 'r')
            self.remote_pid = pid

            # close unused fds
            os.close(child_stdout_in)
            os.close(child_stdin_out)

            # sigchld handler
            signal.signal(signal.SIGCHLD, sigchld_handler)

            # renice the player
            os.system("renice -5 -g %d > /dev/null 2>&1" % self.remote_pid)

            multiplex.add_watch(self.remote_from, multiplex.POLLIN, self.hook)

            fcntl.fcntl(self.remote_from, FCNTL.F_SETFL, os.O_NONBLOCK)
            fcntl.fcntl(self.remote_to, FCNTL.F_SETFL, os.O_NONBLOCK)

        self.status = "playing"
        ui.refresh()

    def _rsnd(self, msg, expect=None, func=None):
        """ send a command to a mpg123 remote. adds \n """

        # if the latest status output wasn't processed, try read()

        if self.remote_pid==None:
            debug("in_mpg123/rsnd: player not running")
            return

        count = 0
        while len(self.expect_stack):
            # make sure an ack from a previous cmd hasn't come in
            # FIXME: this is evil, and slow, seeing as mpg321 now supports
            # asyncronous commands. changing it would break compatibilty with
            # mpg123 though, so i'll have to think about moving this into a
            # seperate module if i do that.
            self.drain_input()

        os.write(self.remote_to, msg + "\n")

        self.expect_stack.append((expect,func))

        return 1

    def drain_input(self):
        """ read as much as we can from mpg123's output, until its empty. """

        while 1:
            try:
                line = self.remote_msg.readline().strip()
            except (OSError, IOError), e:
                if e[0] == 11:
                    return
                else:
                    debug("line/read oserr")
                    return

            if line=="": # when there's nothing to read
                return

            # i fear that even though this is above the normal verbosity
            # level, it'll chew up CPU.
            #debug("recvd: " + str(line), 3)

            # now process the line

            # if this output was expected
            if len(self.expect_stack):
                if self.expect_stack[0][0]==None:
                    self.expect_stack.pop(0)
                    return

                if line.find(self.expect_stack[0][0])==0:
                    if self.expect_stack[0][1]:
                        self.expect_stack[0][1](line)
                    self.expect_stack.pop(0)
                    return

            if self._parse_status(line):

                # normally update the status display, and re-schedule
                # us to receive events on the FD

                ui.new_time()
                ui.refresh()

            else:
                # not a usual status line, check what's going on
                self._parse_other(line)


    def filename_to_meta(self, path):
        """ given a file pathname, return the artist, title etc """

        ret = {}
        try:
            f = mp3meta.parse_filename(os.path.basename(path))
            if  not f:
                f = mp3meta.parse_id3(path)
                ret['artist'] = mp3meta.nice_case(f['artist'])
                ret['title'] = mp3meta.nice_case(f['title'])
            else:
                ret['artist'] = mp3meta.nice_case(f['artist'])
                ret['title'] = mp3meta.nice_case(f['title'])

        except: pass

        if not ret:
            ret['artist'] = os.path.basename(path).replace(".mp3", "")
            ret['title'] = ""

        return ret

    def hook(self, fd):
        """this function is called when there's action on our current FD."""

        self.drain_input()

    def on_sigchild(self):
        pass

    def _finish(self):
        """ song has finished, clean up, update the user interface, etc """

        self.playing = 0
        self.stop()

        # signal to the player that we need advance
        muse.player.finished()

    def _parse_other(self, line):
        """ check for other things, like a song finishing abruptly etc """

        if line.find("@P 0")==0:
            #debug("finish due to @p 0")
            self._finish()

        elif line.find("@E")==0:
            pass
            #debug("error reported from remote!")

        elif line.find("@S")==0:
            pass

        elif line.find("@R MPG123")==0:
            pass

        elif line.find("@I")==0:
            pass

        else:
            debug("unhandled: " + str(line))

    def _parse_status(self, line):
        """takes a line of output from mpg123 and fill in the status info
        returns true on success."""

        m = self.status_match.search(line)

        if m:
            # a successful status line from mpg123
            self.frame      = int(m.group(1))
            self.rem_frame  = int(m.group(2))
            self.cur_secs  = float(m.group(3))
            self.rem_secs   = float(m.group(4))
            # processed successfully
            return 1

        else:
            # not processed
            return 0

    def is_playing(self):
        return self.playing

    def validate_filename(self, file):
        """given a filename, returns if it is able to parse this file"""

        if self.file_match.search(file):
            # the file is one of ours
            return 1
        else:
            # the file isn't ours
            return 0

    def validate_content(self, file):
        """given a filename, checks content - is it able to parse file?"""

        # this gets a bit trickier. using the mp3.py module, we
        # attempt to parse the file. if we catch an exception, chances
        # are this isn't an mp3 file.

        debug("broken atm"); assert(0)

        f = mp3meta.file(file)

        if f.is_valid():
            # this file *appears* to be an mp3 file
            return 1
        else:
            # whoops, must be another file
            return 0

    def locked(self):
        """ we're locked if there's no current file to play """
        return self.file == None

    def play(self, file=None, frame=0):
        """plays a filename. """
        self._play(file, frame)

    def _play(self, file=None, frame=0, player=None):

        if self.paused:
            return

        self.paused = 0
        self.file = file
        self.playing = 1

        self.song = os.path.basename(file)

        #        if not self.remote_to: # FIXME: detect if process alive
        #            self.start_remote()

        if not self.remote_to:
            debug("no remote on play!")
            return

        # we match on @F 1, not @I, as mpg123 doesn't seem to send @I when
        # playing the same song twice?!?
        self._rsnd("load " + self.file, "@F", self.on_play)

    def on_play(self, line):
        """ called when the remote acks the play """
        self.status = "playing"
        ui.new_status()
        ui.refresh()

    def _stop(self):
        """stops any current playing, may be called by pause too"""

        if self.paused:
            return

        if self.is_playing():
            self._rsnd("stop", "@P 0", self.on_stop)

        self.file=None
        self.playing=0

    def stop(self):
        """ calls _stop and then resets everything, including user interface """

        if self.paused:
            return

        self._stop()

        self._reset_state()


    def on_stop(self, line):
        self.status = "stopped"
        ui.new_status()
        self.secs = None
        ui.new_time()
        ui.refresh()

    def _reset_state(self):
        self.paused = 0
        self.file = None
        self.frame = 0
        self.rem_frame = 0
        self.skip_destination=0
        self.cur_secs = 0
        self.rem_secs = 0

        multiplex.del_timer('skip')


    def pause(self):
        """stop the current song playing, but remember where it is"""

        if self.paused == 1:
            # if we're currently paused, unpause
            self._rsnd("pause", "@P 2")
            self.status = "playing"
            ui.new_status()
            ui.refresh()
            self.paused = 0

        elif not self.locked():
            # we're not paused - pause.
            self._rsnd("pause", "@P 1")
            self.status = "paused"

            ui.new_status()
            ui.refresh()
            self.paused = 1

    def fwd(self):
        """jump forward in a track"""

        # rewrote this to have a set delay before skipping - works better
        # on slower machines.

        # we check here, but we must also check in the hook

        if not self.is_playing(): return

        if (self.skip_destination+200) > self.rem_frame:
            self.skip_destination = 0
        else:
            self.skip_destination += 200

        secs = self.skip_destination / 38

        ui.notice("skipping forward ~ %.2d:%.2d" % (secs/60, secs%60))
        ui.refresh()

        multiplex.set_timer("skip", 0.5, lambda s=self: s._fwd())

    def _fwd(self):

        if not self.is_playing():
            self.skip_destination = 0
            return

        self._rsnd("jump +" + str(self.skip_destination))
        ui.notice("")

        self.skip_destination = 0

    def back(self):
        """jump back in a track"""

        if not self.is_playing():
            self.skip_destination = 0
            return

        if self.frame - (self.skip_destination + 200) < 0:
            self.skip_destination = 0
        else:
            self.skip_destination += 200

        secs = self.skip_destination/38 # rough, very rough! :)

        ui.notice("skipping back ~ %.2d:%.2d" % (secs/60, secs%60))
        ui.refresh()

        multiplex.set_timer("skip", 0.5, lambda s=self: s._back())

    def _back(self):

        if not self.is_playing():
            self.skip_destination = 0
            return

        self._rsnd("jump -" + str(self.skip_destination))
        ui.notice("")

        self.skip_destination = 0

    def message(self, msg):

        if msg == "switch_player":
            if self.mpg123_executable=="mpg123":
                ui.notice("switching to mpg321 (madplay)")
                ui.refresh()
                self.mpg123_executable="mpg321"
                self.stop_remote()
                self.start_remote()
            else:
                ui.notice("switching to mpg123 (standard)")
                ui.refresh()
                self.mpg123_executable="mpg123"
                self.stop_remote()
                self.start_remote()


    def time_seek(self, time):
        # determine frame to jump to:
        total = self.frame + self.rem_frame
        dest = total * (time / self.time_total())

        # compensate for seeking bug in mpg123
        if (dest+32) > total:
            dest = total-32

        self._rsnd("jump %d" % dest)

    def time_tell(self):
        return self.cur_secs

    def time_total(self):
        return self.cur_secs + self.rem_secs


