#!/usr/bin/env python
#
# multiplex.py - provides an interface to the main loop, where modules can
#                provide fds to watch, etc. this module also provides a timer
#                interface.
#
# started 8/8/01 by Damien Elmes <resolve@repose.cx>
# $Id: multiplex.py,v 1.1.1.1 2003/02/11 06:07:41 resolve Exp $
#


import sys
import time

# we import select in order to provide stuff like POLLIN
from debug import *
from select import *

poll_table = {}
timers = []
idles = []
running = 1

# sets the resolution of timers. 1sec is a safe choice, as time.time() is only
# accurate to that on some systems. clock() could be used to implement a timer
# with a higher resolution, although it's always going to stuff up if one of
# the hooks takes a long time.
#
# scratch that, using 0.05sec :-)
idle_granularity = 50

def add_watch(fd, mask, hook):
    """ Add an entry to the watch table.

    Mask takes the same values as select.poll.register(..., mask) """

    poll_table[fd] = (mask, hook)

def del_watch(fd):
    """ Delete an entry from the poll table """

    del poll_table[fd]

def add_timer(label, secs, hook):

    target = time.time() + secs

    timers.append({ 'label': label, 'target': target,
                    'time': secs, 'hook': hook })

def set_timer(label, secs, hook):
    """ call hook when secs have elapsed - label is used for managing the
    timer. if label is already used, then update the details. """

    for unit in timers:
        if label == unit['label']:
            unit['target'] = time.time() + secs
            unit['hook'] = hook
            return

    # not existing, add new onen
    timers.append({ 'label': label, 'target': time.time() + secs, 'hook': hook,
                    'time': secs })

def del_timer(label):
    """ delete a timer - this should no-op if the timer has been called"""

    def filt(unit, label=label):
        if label == unit['label']:
            return 0
        return 1

    global timers
    timers = filter(filt, timers)

def add_idle(func):
    """ add an idle task to be called. if the task returns 0, it will be
    removed """

    #print "idle function added: %s" % `func`
    idles.append(func)

def main_loop():
    """ poll and call any hooks where appropriate. generally you'd call this
    from your main function. this function also checks to make sure to make
    sure there are no outstanding timers waiting. """

    while running:

        # setup a poll object with our current list of FDs
        polls = poll()

        for x in poll_table.items():
            (fd, (mask, _)) = x
            polls.register(fd, mask)

        try:
            pending_fds = polls.poll(idle_granularity)
        except error:
            debug("select.error on poll, this is probably a resized xterm.")
            pending_fds = []

        # iterate through all the fds with action on em.
        for x in pending_fds:
            (fd, mask) = x
            poll_table[fd][1](fd)


        # check to see if any timers need to be called
        cur_time = time.time()

        # if a timer's expired, run it, and if it returns a boolean false,
        # remove it from the list of timers.
        def filt(t, cur_time=cur_time):
            if t['target'] < cur_time:
                ret = t['hook']()
                if ret:
                    # keep it, but we'll have to refresh target
                    t['target'] = time.time() + t['time']
                    return 1
                else:
                    return 0
            return 1

        global timers
        timers = filter(filt, timers)

        # nothing's doing, let's run idle tasks
        # run any idle tasks, ones that return 0 are removed
        global idles
        idles = filter(apply, idles)
