;;; muse.el ---

;; Copyright (C) 2003  Damien Elmes <resolve@repose.cx>

;; Author: Damien Elmes <resolve@repose.cx>
;; Keywords:

;; This file is free software; you can redistribute it and/or modify
;; it under the terms of the GNU General Public License as published by
;; the Free Software Foundation; either version 2, or (at your option)
;; any later version.

;; This file is distributed in the hope that it will be useful,
;; but WITHOUT ANY WARRANTY; without even the implied warranty of
;; MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
;; GNU General Public License for more details.

;; You should have received a copy of the GNU General Public License
;; along with GNU Emacs; see the file COPYING.  If not, write to
;; the Free Software Foundation, Inc., 59 Temple Place - Suite 330,
;; Boston, MA 02111-1307, USA.

;;; Commentary:

;;


;;; Code:

(setq muse-host "reflex"
      muse-port "8008")

(defun muse-process ()
  "Returns a muse process, creating one if necessary"
  (let* ((buf (muse-buffer))
         (proc (get-buffer-process buf)))
    (if (or (null proc) (eq (process-status proc) 'failed))
        (muse-create-process buf)
      (get-buffer-process buf))))

(defun muse-buffer ()
  (let ((buf (get-buffer "*muse*")))
    (unless buf
      (setq buf (get-buffer-create "*muse*"))
      (muse-interaction-mode))
    buf))

(defun muse-interaction-mode ()
  (with-current-buffer buf
    (setq major-mode 'muse-interaction-mode)
    (local-set-key (kbd "/") 'muse-list)
    (local-set-key (kbd "l") 'muse-go-back)
    (local-set-key (kbd "c") 'muse-cmd)
    (local-set-key (kbd "p") 'muse-play-next)
    (local-set-key (kbd "RET") 'muse-follow-line)
    (local-set-key (kbd "M-RET") 'muse-play-current)
    (setq muse-current-path "")
    (setq header-line-format '(:eval (concat " (muse) " muse-current-path)))))

(global-unset-key (kbd "C-c C-SPC"))
(global-set-key (kbd "C-c C-SPC") 'muse-list-all)

(defun muse-create-process (buf)
  (let ((proc (open-network-stream "muse" buf muse-host muse-port)))
    (set-process-filter proc nil)
    proc))

(defun muse-cmd (str)
  (interactive (list (read-from-minibuffer "Muse command: ")))
  (process-send-string (muse-process) (concat str "\n")))

(defmacro define-line-filter (filter args &rest bodyforms)
  (let ((collection (gensym))
        (str (nth 1 args)))
    `(defun ,filter (,@args)
       (if (string-match "\n" ,str)
           (progn
             (setq ,str (and (setq s
                                  (if (boundp (quote ,collection))
                                      (progn
                                        (setq ,str (concat ,collection ,str))
                                        (setq ,collection nil)
                                        ,str)
                                    ,str))
                            (substring s 0 (1- (length s)))))
             ,@bodyforms)
         (if (boundp (quote ,collection))
             (setq ,collection (concat ,collection ,str))
           (setq ,collection ,str))))))

(defun muse-match (str callback)
  (lexical-let ((callback callback))
    (define-line-filter muse-get-matches (proc st)
      (setq muse-matches (read st))
      (set-process-filter (muse-process) 'muse-process-filter)
      (funcall callback)))
  (set-process-filter (muse-process) 'muse-get-matches)
  (process-send-string (muse-process) (concat "llist " str "\n")))

(defun muse-list-all ()
  (interactive)
  (setq muse-current-path "")
  (muse-match muse-current-path 'muse-insert-matches))

(defun muse-list (spec)
  (interactive (list (read-from-minibuffer "(m) list: ")))
  (muse-match spec 'muse-insert-matches))

(setq muse-match-history nil)

(define-line-filter muse-process-filter (proc str)
  (message "Muse: %s" str))

(defun muse-add-and-play (path)
  (process-send-string (muse-process)
                       (concat "p add " path "\n")))

(defun muse-insert-matches (&optional arg)
  (switch-to-buffer (muse-buffer))
  (if (eq 1 (length muse-matches))
      (progn
        (muse-add-and-play (muse-path-add (car muse-matches)))
        nil)
    (erase-buffer)
    (dolist (l muse-matches)
      (insert l "\n"))
    (sort-lines nil (point-min) (point-max))
    (goto-char (point-min))
    t))

(defun muse-path-add (p)
  (concat muse-current-path "/" p))

(defun current-line ()
  (buffer-substring (line-beginning-position)
                    (line-end-position)))

(defun muse-follow-line ()
  (interactive)
  (muse-follow (current-line)))

(defun muse-play-current ()
  (interactive)
  (muse-add-and-play (muse-path-add (current-line))))

(defun muse-play-next ()
  (interactive)
  (process-send-string (muse-process) "play\n"))

(defun muse-follow-absolute (str)
  (lexical-let* ((newstr str))
    (message "following %s" str)
    (muse-match str
                (lambda ()
                  (when (muse-insert-matches)
                    (setq muse-current-path newstr))))))

(defun muse-follow (str)
  (if (eq (aref str 0) ?/)
      (muse-follow-absolute str)
    (muse-follow-absolute (muse-path-add str))))

(defun muse-go-back ()
  (interactive)
  (let ((newpath (replace-regexp-in-string "/[^/]+$" "" muse-current-path)))
    (when (string= newpath muse-current-path)
      (error "Can't go back any further!"))
    (muse-follow-absolute newpath)))

(copy-face 'mode-line 'header-line)

(provide 'muse)
;;; muse.el ends here

