#!/usr/bin/env python
#
# $Id: orbital.py,v 1.1.1.1 2003/02/11 06:07:41 resolve Exp $
#
# Started 5/8/01 by Damien Elmes <resolve@repose.cx>
# This file is licensed under the GPL.
#

"""
Interface to a Matrix Orbital screen and optional keypad
"""

import os, sys
import atexit
import termios
if sys.version_info[0] >= 2 and sys.version_info[1] >= 1:
    import termios as TERMIOS
else:
    import TERMIOS

import curses
import multiplex
from debug import *

ESC = chr(0xFE)

CUSTOM_CHAR_LIMIT=8

class error:
    def __init__(self, value=""):
        self.value = value

    def __str__(self):
        return repr(self.value)

class pending:
    def __init__(self, value=""):
        self.value = value

    def __str__(self):
        return repr(self.value)

class Orbital:
    """ create a new orbital connection """

    def _onexit(self):
        """ called by atexit, clean up the screen and prevent further
        modifications """

        # reset serial
        termios.tcsetattr(self.screen_fd, TERMIOS.TCSANOW, self.old_termios)

    def __init__(self, device, baud=19200):

        baud = "B" + str(baud)

        # FIXME: raise orbital.pending
        try:
            self.screen_fd = os.open(device, os.O_RDWR | os.O_NOCTTY)
        except OSError, e:
            raise error, str(e)

        # set speed
        try:
            self.old_termios = termios.tcgetattr(self.screen_fd)
        except termios.error, e:
            raise error, str(e)

        atexit.register(self._onexit)

        old = self.old_termios[:]

        # control flags
        old[2] |= TERMIOS.CS8 | TERMIOS.CLOCAL

        # iflags: don't want terminal echo etc
        old[3] &= ~(TERMIOS.ECHO | TERMIOS.ECHONL | \
                    TERMIOS.ICANON | TERMIOS.ISIG)

        #old[2] &= ~(TERMIOS.CSIZE|TERMIOS.PARENB|TERMIOS.HUPCL);

        old[1] &= ~TERMIOS.OPOST;

        #old[0] &= ~(TERMIOS.IGNBRK|TERMIOS.BRKINT|TERMIOS.PARMRK| \
        #            TERMIOS.ISTRIP|TERMIOS.INLCR|TERMIOS.IGNCR|TERMIOS.ICRNL|\
        #            TERMIOS.IXON);

        old[4] = getattr(TERMIOS, baud) # ispeed
        old[5] = getattr(TERMIOS, baud) # ospeed

        termios.tcsetattr(self.screen_fd, TERMIOS.TCSANOW, old)

        # set up a hook for incoming data
        mask = multiplex.POLLIN
        multiplex.add_watch(self.screen_fd, mask, self.on_keypad)

        self.keypad_cb = None

    def on_keypad(self, fd):
        if self.keypad_cb != None:
            c = os.read(fd, 1)
            self.keypad_cb(c)

    def set_keypad_cb(self, cb):
        self.keypad_cb = cb


    def write(self, msg, x=None, y=None, size=None, pad=None):
        """ write a message to the screen

        x and y coods are optional, as is size, which the message will be
        chopped to. """

        if x != None and y != None:
            self.set_pos(x, y)

        if size != None:
            msg = msg[:size]

        if pad != None and size!=None:
            msg += " " * (size-len(msg))

        os.write(self.screen_fd, msg)

    def clear_screen(self):
        self.write(ESC + 'X')

    def set_pos(self, x, y):
        """ move the cursor to (x, y) """

        self.write(ESC + 'G' + chr(x) + chr(y))

    def set_startup_screen(self, msg):
        """ set the new screen on startup """

        # crop or pad
        if len(msg) < 80:
            msg += " " * (80-len(msg))
        elif len(msg)>80:
            msg = msg[:80]

        self.write(ESC + '@' + msg)

    def set_contrast(self, contrast):
        if 0 > contrast > 255:
            raise TypeError, "Contrast must be between 0-255!"

        self.write(ESC + 'P' + chr(contrast))

    def backlight_on(self):
        self.write(ESC + 'B')

    def backlight_off(self):
        self.write(ESC + 'F')

    def set_custom_char(self, index, bytes):
        """ write a custom char at index (from 0-7), using a seq of 8 bytes

        the bytes only cares about the lower 5 bits - please see orbital
        documentation """

        if index > CUSTOM_CHAR_LIMIT+1:
            raise ValueError, "Index exceeds the number of custom chars screen"\
                  " can store"

        if len(bytes) != 8:
            raise ValueError, "Bytestring should be exactally 8 bytes"

        self.write(ESC + 'N' + chr(index))
        for x in bytes:
            self.write(chr(x))

    def clear_keybuffer(self):
        """ flush the key buffer, discarding output """
        self.write(ESC + 'E')


class OrbitalTester(Orbital):
    """ A framework for testing code that interfaces with Matrix Orbital
    screens. """

    def __init__(self, device=None, baud=19200):
        """ We ignore any information passed to us as we don't use it """

        self.code_table = {
            'G': self.on_set_pos,
            'X': self.on_clear,
##             '@': self.on_set_startup,
##             'P': self.on_set_contrast,
##             'B': self.on_set_backlight,
##             'F': self.on_set_backlight,
            'N': self.on_set_custom_char,
##             'E': self.on_key_buffer_flush,
            }

        self.buffer = ""
        self.max_x = 40
        self.max_y = 2
        self.keypad_cb = None

        self._stdscr = curses.initscr()

        self._stdscr.addstr(1, 10, "Matrix Orbital emulation")
        self._stdscr.refresh()
        self.stdscr = curses.newwin(2, 40, 3, 3)

        multiplex.add_watch(sys.stdin.fileno(), multiplex.POLLIN,
                            self.on_keypad)

        curses.noecho()
        curses.cbreak()
        self.stdscr.keypad(1)
        self.stdscr.leaveok(1)
        curses.curs_set(0)
        self.stdscr.clear()

        # setup exit handler
        atexit.register(self._onexit)

        # now lay out the screen as it should appear

        self.stdscr.move(0, 0)
        self.stdscr.attron(curses.A_REVERSE)

    def on_clear(self, msg):
        self.stdscr.clear()

        return 0

    def on_set_pos(self, msg):
        if len(msg) < 2:
            raise pending

        (x, y) = (ord(msg[0]), ord(msg[1]))

        if x > self.max_x: raise "x out of bounds"
        if y > self.max_y: raise "y out of bounds"

        self.stdscr.move(y-1, x-1)

        return 2

    def on_set_custom_char(self, msg):
        if len(msg) < 9:
            raise pending

        return 9

    def write(self, msg, x=None, y=None, size=None, pad=None):
        """ write a message to the screen

        x and y coods are optional, as is size, which the message will be
        chopped to. """

        if x != None and y != None:
            self.set_pos(x, y)

        if size != None:
            msg = msg[:size]

        if pad != None and size!=None:
            msg += " " * (size-len(msg))

        self.buffer += msg

        self._parse()

    def _parse(self):
        while self.buffer:
            #debug("buffer: %s" % `self.buffer`)
            try:
                self._parse_token()
            except pending:
                break

        self.stdscr.refresh()

    def _parse_token(self):
        """ Grab a token, parse it, and return what's left. """

        # if it's just text, output to the screen at our current location.
        # else if it's a command character, handle that. we return a new
        # string minus what we've processed.

        for i in range(len(self.buffer)):
            if self.buffer[i] == ESC:
                idx = self._handle_control(self.buffer)
                self.buffer = self.buffer[idx:]
                return
            else:
                c = self.buffer[i]

                if ord(c) in range(8):
                    c='+'
                try:
                    self.stdscr.addch(c)
                except curses.error:
                    pass
                self.buffer =  self.buffer[i+1:]
                return

    def _handle_control(self, msg):
        if len(msg) < 2:
            raise pending

        code = msg[1]
        msg = msg[2:]
        return self.code_table[code](msg) + 2


    def _onexit(self):
        """ called by atexit, clean up the screen and prevent further
        modifications """

        curses.endwin()


    def on_keypad(self, fd):
        """ We have console input. Emulate orbital events """

        # keymap = {
        #     "R": 1,      # key layout:
        #     "Q": 2,      #
        #     "P": 3,      #    1 2 3
        #     "M": 4,      #    4 5 6
        #     "L": 5,      #    7 8 9
        #     "K": 6,      #    * 0 #
        #     "H": 7,      #
        #     "G": 8,      #    3 6 9 #
        #     "F": 9,      #    2 5 8 0
        #     "B": 0,      #    1 4 7 *
        #     "C": '*',    #
        #     "A": '#'     #
        #     }

        if self.keypad_cb != None:
            c = os.read(fd, 1)
            if ord(c) == 59: c = "R"
            elif ord(c) == 97: c = "Q"
            elif ord(c) == 39: c = "P"
            elif ord(c) == 113: c = "M"
            elif ord(c) == 111: c = "L"
            elif ord(c) == 44: c = "K"
            elif ord(c) == 106: c = "H"
            elif ord(c) == 101: c = "G"
            elif ord(c) == 46: c = "F"
            elif ord(c) == 107: c = "B"
            elif ord(c) == 117: c = "C"
            elif ord(c) == 112: c = "A"
            else: return

            self.keypad_cb(c)
