#!/usr/bin/env python
#
# playlist.py - a class
#               playlist instances
#
# $Id: playlist.py,v 1.1.1.1 2003/02/11 06:07:41 resolve Exp $
#

# we provide two playlist instances for use in muse
# these are the 'list' - a list of all files available for playing
# and the 'queue', a list of files which have been queued up. when the queue
# is exhausted, playing will resume at the current playlist position.

# NB: these are defined after the playlist class

#from __future__ import nested_scopes

import os
import pickle
import re
import string
from stat import *
from debug import *

class error:
    """Raised when there is an error relating to a playlist"""

    def __init__(self, value):
        self.value = value

    def __str__(self):
        return repr(self.value)

class Item(dict):
    def __init__(self, d={}):
        dict.__init__ (self, d)
        s = os.path.basename(self['path'])
        self['song'] = re.sub (".([mM][pP]3|[oO][gG][gG])$", "", s)

        l = string.split(self['song'], "-", maxsplit=1)
        if len(l) < 2:
            # that didn't work, try alternative split s trategy
            if self['song'][0] == "(":
                f = self.find_matching_paren (1, self['song'])
                self['artist'] = self['song'][1:f]
                self['title'] = self['song'][f+2:]
            else:
                self['artist'] = self['song']
                self['title'] = ""
        else:
            self['artist'] = l[0].strip()
            self['title'] = l[1].strip()

        self['song'] = self['artist'] + " - " + self['title']

    def find_matching_paren(self, index, str):
        count=1
        for c in range(len(str)):
            cc = c + index
            if str[cc] == ")":
                count -= 1
            elif str[cc] == "(":
                count += 1
            if count == 0:
                return cc
        return None

class Playlist(list):
    callback_types = ("list", "selected", "playing")
    def __init__(self, seq=[], history=[]):
        list.__init__(self, seq)

        self.locked = 0

        self._callbacks = {}
        self._pending = {}

        for i in Playlist.callback_types:
            self._callbacks[i] = []
            self._pending[i] = 0

        self._selected = None
        self._playing = None

        if history != None:
            if not history:
                self.history = []
            else:
                self.history = history

    def append(self, obj):
        list.append(self, obj)
        self.update("list")

    def insert (self, pos, obj):
        list.insert(self, pos, obj)
        if pos <= self._selected:
            self._selected += 1
        self.update("list")

    def remove(self, obj):
        list.remove(self, obj)
        self._update_history(obj)
        self.update("list")

    def clear (self):
        list.__init__(self)
        self._selected = None
        self.update ("list")

    def clear_history (self):
        self.history = []
        self.update ("list")

    def update(self, type):
        """Run `types's hooks, or flag them for running."""
        if self.locked:
            self._pending[type] = 1
        else:
            self._pending[type] = 0
            for c in self._callbacks[type]: c()

    def _update_history(self, entry):
        if getattr(self, 'history')!=None:
            self.history.append(entry)

    def add_hook(self, type, hook):
        """ add a hook to be called when when type is marked as update """
        if type not in Playlist.callback_types:
            raise ValueError, "type not in callback_types"

        self._callbacks[type].append(hook)

    def get_selected (self):
        return self._selected
    def set_selected (self, s):
        self._selected = s
        self.update('selected')
    selected = property (get_selected, set_selected, None,
                         "The currently selected song")

    def get_playing (self):
        return self._playing
    def set_playing (self, s):
        self._playing = s
        self.update('playing')
    playing = property (get_playing, set_playing, None,
                        "The currently playing song")

    def freeze(self):
        self.locked += 1

    def thaw(self):
        self.locked -= 1
        if self.locked < 0:
            raise error, "Can't thaw when not frozen!"
        elif self.locked == 0:
            for (k,v) in self._pending.items():
                if v == 1:
                    self.update(k)

    def __delitem__ (self, item):
        self._update_history(self[item])
        list.__delitem__ (self, item)
        self.update('list')


    def advance (self):
        """Take action when the playing song has finished. Remove the playing
        song from the queue, and ensure selected is valid"""
        self.freeze()
        if self.selected != None:
            del self[self.selected]
        l = len(self)
        if l > 0:
            if self.selected == l or \
               self.selected == None:
                self.selected = l - 1
        else:
            self.selected = None
            self.selected_end()
        self.thaw()
        return self.selected

    def selected_end(self):
        # if the playlist isn't empty, jump back to the start
        print "Playlist is now empty"

    def add_uri(self, uri):
        """ Add a file, directory or m3u. """

        if uri.find(".m3u", -4) >= 0:        #FIXME case insensitive
            self.add_m3u(uri)
        elif os.path.isdir(uri):
            self.add_directory_recursively(uri)
        else:
            self.add_file(uri)

    def add_m3u(self, file):
        """ read a .m3u specifying a list of mp3s to play

        returns a list of the added files. """

        f = open(file).readlines()

        files = []
        i = 1
        for x in f:
            x = x.strip()
            songfile = os.path.join(os.path.dirname(file), x)
            self.add_file(songfile, {'playlist': file, 'track': i})
            files.append(songfile)
            i += 1

        return files

    def add_directory_recursively(self, dir):
        """ recurse through and add all files that have a handler """
        self.freeze()
        os.path.walk(dir, self._add_recursive, None)
        self.thaw()

    def _add_recursive(self, arg, dir, files):

        m3u_files = []
        pendingfiles = []

        for file in files:
            file = os.path.join(dir, file)

            if os.path.isfile(file):
                if file.endswith(".m3u"):
                    m3u_files.extend(self.add_m3u(file))
                    m3u_files.append(file)
                else:
                    pendingfiles.append(file)
            elif os.path.islink(file) and os.path.isdir(file):
                os.path.walk(file, self._add_recursive, None)

#        pendingfiles.sort()
        for file in pendingfiles:
            if file in m3u_files:
                continue
            self.add_file(file)

        return 0

    def add_file(self, path, extra=None):
#         if not os.path.isfile(path):
#             debug("fileops: can't read; dropping " + str(path))
#             return

        #         if not muse.player.have_handler(path):
        #             debug("dropping unhandled file type:\n " + \
        #                   os.path.basename(path), 4)
        #             return

        #t = muse.player.filename_to_meta(path)

        dict = {
            'path':     path,
            'artist':   'artist', #t['artist'],
            'title':    'title' # t['title']
            }

        if extra:
            dict.update(extra)

        self.append (Item(dict))

    def add_cache (self, cache):
        self.freeze()
        f = cache['files']
        if f:
            self._add_recursive(None, cache['path'], f)
        for i in cache['dirs'].values():
            self.add_cache(i)
        self.thaw()

class FSCacheFactory:
    def __init__ (self):
        pass

    def update (self, path, cache={}):
        if not os.path.exists(path):
            return {}
        mtime = cache.get("mtime", 0)
        dirs = cache.get("dirs", {})
        files = cache.get("files", [])
        newtime = os.stat(path)[ST_MTIME]

#        print "p, p = %s %s" % (path, cache['path'])
        if newtime > mtime or cache.get("path", "") != path:
            # our cache is dirty
            files = []
            print "unclean: " + path
            newdir={}
            for elem in os.listdir(path):
                fullp = os.path.join(path, elem)
                if os.path.isdir(fullp):
                    d = self.update(fullp, dirs.get(elem, {}))
                    if d: newdir[elem] = d
                else:
                    files.append(fullp)
            dirs=newdir
        else:
            newdir={}
            for (elem, ent) in dirs.items():
                fullp = os.path.join (path, elem)
                d = self.update(fullp, ent)
                if d: newdir[elem] = d
            dirs=newdir


        cache["dirs"] = dirs
        cache["files"] = files
        cache["mtime"] = newtime
        cache["path"] = path

        return cache

    def load (self, f, path=None):
        try:
            ff = open (f)
            up = pickle.load (ff)
        except IOError:
            up = {}
        if not path:
            path = up['path']
        up = self.update (path, up)
        return up

    def save (self, f, cache):
        pickle.dump (cache, open (f, "w"))

if __name__ == '__main__':
    print "Starting tests.."

    p = Playlist()

    ff = FSCacheFactory()

    p.add_file ("foo")
    p.clear_history()

    p.clear()

    print `p`

