#!/usr/bin/env python
#
# playlist.py - provides a playlist class, a simple media object, and two
#               playlist instances
#
# $Id: playlist_hack.py,v 1.1.1.1 2003/02/11 06:07:41 resolve Exp $
# started 17/8/01 by Damien Elmes <resolve@repose.cx>
#

# we provide two playlist instances for use in muse
# these are the 'list' - a list of all files available for playing
# and the 'queue', a list of files which have been queued up. when the queue
# is exhausted, playing will resume at the current playlist position.

# NB: these are defined after the playlist class

from __future__ import nested_scopes

import os
import types

from UserDict import UserDict
from UserList import UserList
from debug import *

class error:
    """Raised when there is an error relating to a playlist"""

    def __init__(self, value):
        self.value = value

    def __str__(self):
        return repr(self.value)

class Item(UserDict):
    """ while a playlist can contain arbitary objects, muse uses items in
    place, which take care of providing empty values for predefined non-valid
    attributes, such as track numbers on files not associated with a
    playlist """

    def __init__(self, dict={}):
        """ Initialise an item. Takes a dict. """

        self.data = dict
        id = None

    def __getattr__(self, attr):
        """ Get an attribute.
        We proxy requests here for standard playlist stuff. """

        import inspect
        c1 = inspect.getouterframes(inspect.currentframe())[1][1]
        c2 = inspect.getouterframes(inspect.currentframe())[1][3]
#        debug( "getattr %s: %s: %s" % (attr, c1,c2))

        try:
            return UserDict.__getattr__(self, attr)
        except AttributeError:
            try:
                return self.__getitem__(attr)
            except KeyError, e:
                raise AttributeError, e

        raise AttributeError

    def __getitem__(self, item):
        """Proxy requests for some attributes which we calculate in real
        time."""

        if self.data.has_key(item):
            return self.data[item]

        # convenience
        if item == "song":
            return self['artist'] + " - " + self['title']

        if item == "dir":
            return os.path.dirname(self['path'])

        if item == "file":
            return os.path.basename(self['path'])

        if item in ("track",):
            return 0

        if item in ("playlist",):
            return ""

        # default
        raise KeyError, "Requested key %s does not exist" % item


class Playlist(UserList):
    """A playlist is a list of comprised of a number of Items.

    The three pieces we care about here are:
      selected
      playing
      data

    Which point to the selected song, playing song, and the internal list.
    Selected and playing are not set so it is the responsibility of the calling
    code to do that. """

    def __init__(self, update_from=None):
        # this must be set here, first, or things will break - warned you! :)
        self.__dict__['locked'] = 0
        self.__dict__['callback_types'] = ('list', 'selected', 'playing')
        self.__dict__['data'] = []
        self.__dict__['revmap'] = {}

        # create hook lists:
        for i in self.callback_types:
            setattr(self, i+'_callbacks', [])

        UserList.__init__(self, update_from)

        self.__dict__['_selected'] = 0
        self.__dict__['_playing'] = 0

        self.last_id = 0

        self.last_jump = None

    def append(self, item):
        """Add `item' to playlist. This should be an Item object"""

        assert (isinstance (item, Item))

        self.data.append (item)
        item.id = len (self.data) - 1
        self.dirty("list")

    def __getitem__ (self, index):
        pass

    def index(self, obj):
        """ Hackishly look up the ID of an object in our revmap """
        debug ("index() used")

    def dirty(self, type):
        """ mark type as dirty or call the appropriate hooks """

        if type not in self.callback_types:
            raise ValueError, "type not in callback_types"

        if self.locked:
            self.__dict__['_dirty_'+type] = 1
        else:
            self.__dict__['_dirty_'+type] = 0
            [x() for x in getattr(self, type+'_callbacks')]

    def add_hook(self, type, hook):
        """ add a hook to be called when when type is marked as dirty """

        if type not in self.callback_types:
            raise ValueError, "type not in callback_types"

        getattr(self, type+'_callbacks').append(hook)

    def __repr__(self):
        return repr(self.data)

    def __setattr__(self, name, value):
        """ when an instance var is changed, call our hooks """

        if name in ("selected", "playing"):
            if len(self.data)==0: raise error, "Playlist empty"
            if value<0 or value>=len(self.data):
                raise ValueError, "Out of bounds"

            self.__dict__['_'+name] = value
            self.dirty(name)
            return
        elif name == "data":
            self.dirty("list")

        self.__dict__[name] = value

    def __getattr__(self, name):

        if name == "selected":
            return self._selected
        if name == "playing":
            return self._playing
        if name == "sel":
            return self.data[self._selected]
        if name == "pl":
            return self.data[self._playing]

        # if there's a sel_ prefix, pass in the selected index
        if name.find("sel_")==0:
            def apply_with_index(*args):
                return apply(getattr(self, name[4:]), (self._selected,) + args)
            try:
                return apply_with_index
            except AttributeError:
                raise AttriubeError, "attribute %s not found" % name

        # if there's a pl_ prefix, pass in the selected index
        if name.find("pl_")==0:
            def apply_with_index(*args):
                return apply(getattr(self, name[3:]), (self._playing,) + args)
            try:
                return apply_with_index
            except AttributeError:
                raise AttriubeError, "attribute %s not found" % name

        raise AttributeError, "Attribute not found: %s" % name

    def freeze(self):
        """ Disable the calling of playlist hooks while updating.

        This works via ref counts so that you can wrap chunks of code
        and prevent them from updating. """
        self.__dict__['locked']+=1

    def thaw(self):
        """ Enable the calling of playlist hooks while updating """
        self.__dict__['locked']-=1

        if self.__dict__['locked']<0:
            raise error, "Can't thaw when not frozen!"

        if self.__dict__['locked']==0:
            for c in self.callback_types:
                self.dirty(c)

    def jump(self, index, amount=1, wrap=1):
        """ return the index after/before index in a list.

        the playlist will wrap if wrap is true. amount specifies the number of
        elements to move. """

        if len(self.data) == 0:
            raise error, "Playlist empty"

        if amount >= 0:
            dir=1
        else:
            dir=-1

        amount = abs(amount)

        amount = (amount % len(self.data))*dir

        if index+amount >= len(self.data):
            if not wrap:
                raise error, "Wrapped"
            else:
                return (index+amount) - len(self.data)
        elif index+amount < 0:
            if not wrap:
                raise error, "Wrapped"
            else:
                return len(self.data) + (index+amount)
        else:
            return index+amount

    def sort(self):
        """ sort the current playlist """

        pass
#         def by(key, a, b):
#             """ -1 if a < b, 0 ==, 1 if a > b """
#             try:
#                 x = a[key]
#                 y = b[key]
#             except KeyError:
#                 return 0

#             if x == y: return 0
#             if x > y: return 1
#             if x < y: return -1

#         def all(a, b, by=by):
#             foo = by('playlist', a, b)
#             if foo==0:
#                 foo = by('track', a, b)
#                 if foo==0:
#                     foo = by('artist', a, b)
#                     if foo==0:
#                         foo = by('dir', a, b)
#                         if foo==0:
#                             return by('title', a, b)
#                         else: return foo
#                     else: return foo
#                 else: return foo
#             else: return foo

#         self.data.sort(all)
#         self.dirty('list')

    def reverse(self):
        """ reverse the order of the current playlist """
        pass
#         self.data.reverse()
#         self._recache()
#         self.dirty('list')

    def jump_field_changed(self, index, field, direction=1):
        """ search for change in field and set current to the first change """

        if len(self.data) == 0: raise error, "Playlist empty"
        if direction not in (1, -1): raise ValueError, "Direction invalid"

        current = self.data[index][field]
        start = index

        ret=0
        while 1:
            index = self.jump(index, direction)

            if index==start: break

            if self.data[index][field] != current:
                return index

        return -1

    def jump_to_field(self, index, field=None, comp=None, case_sensitive=0):
        """ jump to a particular key in the playlist

        eg, jump_to_field('artist', 'bogart') to jump to 'henry bogart -
        jammin'

        if field and comp are both None (or not passed), then the previous
        search will try to repeat
        """

        if len(self.data) == 0: raise error, "Playlist empty"

        if index >= 0:
            c = index
        else:
            c = index = 1

        # the user would like the previous
        # search repeated
        if field == comp == None:
            if self.last_jump == None:
                #raise error, "No previous search"
                return -1
            (field, comp) = self.last_jump
        else:
            self.last_jump = (field, comp)

        # now see if we match the artist - if we've looped around,
        # then the artist couldn't be found, and we fail.

        # if they've specified case sensitivity, or if comp is not a string,
        # then don't lower() below.
        stringp = type(comp)==type("") and (not case_sensitive)

        while 1:
            c = self.jump(c, 1)
            if c==index:
                return -1

            if stringp:
                if self.data[c][field].lower().find(comp.lower())!=-1:
                    return c
            # if the supplied arg wasn't a string, don't do case-insensitive
            # match, as it makes no sense.
            else:
                if self.data[c][field]==comp:
                    return c

#
# testing
#


if __name__ == '__main__':

    import unittest

    class SimpleTestCase(unittest.TestCase):
        def setUp(self):
            self.p = Playlist()
        def tearDown(self):
            pass
        def testItemCons (self):
            i = Item()
            j = Item({'filename': "foo"})
            assert i != None
            assert j

        def testAppend(self):
            """Test appending information to a playlist

            foo"""
            try:
                self.p.append ("foo")
            except AssertionError:
                passed=1

            if not passed: assert 0

            for i in "foo", "bar", "baz", "quux":
                self.p.append (Item ({'path': i}))

    class IndexingTestCase (unittest.TestCase):
        def setUp (self):
            self.p = Playlist()

            self.list = ("foo", "bar", "baz", "quux")
            self.pl = []

            for i in self.list:
                t = Item ({'path': i})
                self.p.append (t)
                self.pl.append (t)

        def testGet (self):
            """Test inserting/removing items from a playlist"""
            # while we don't have to guarantee ids returned will be
            # consecutive, they must be able to return the correct
            # playlist item
            for i in range (len (self.list)):
                id = self.pl[i].id
                #assert self.list[i] == self.p[id]['path']


    unittest.main()
