#!/usr/bin/env python

import curses
import sys
import os # FIXME, can be removed later
import atexit

import ui
import multiplex
import muse

from debug import debug

PLUGIN_TYPE="ui"

class new(ui.plugin):
    """ provides an abstract class for a interface plugin, for muse"""

    def _onexit(self):
        """ called by atexit, clean up the screen and prevent further
        modifications """
        self.stdscr.clear()
        self.stdscr.keypad(0)
        self.stdscr.leaveok(0)
        curses.nocbreak()
        curses.echo()
        curses.endwin()

        # evil hack to prevent modules from writing to the screen after we close
        ui.refresh = lambda: 1

    def __init__(self):
        ui.plugin.__init__(self)

        self.plugin_type = "ui"
        self.plugin_name = "curses"
        self.plugin_desc = "ncurses interface"

        self.stdscr = curses.initscr()

        curses.noecho()
        curses.cbreak()
        self.stdscr.keypad(1)
        self.stdscr.leaveok(1)
        self.stdscr.clear()

        # setup exit handler
        atexit.register(self._onexit)

        # now lay out the screen as it should appear

        # clear top line
        self.stdscr.move(0, 0)
        self.stdscr.attron(curses.A_REVERSE)
        self.stdscr.addstr(" " * curses.COLS)
        # clear bottom line
        self.stdscr.move(curses.LINES - 1, 0)
        self.stdscr.addstr(" " * (curses.COLS - 1))
        self.stdscr.attroff(curses.A_REVERSE)

        fd = sys.stdin.fileno()
        mask = multiplex.POLLIN
        multiplex.add_watch(fd, mask, self.hook)

        muse.plist.add_hook('selected', self.draw_playlist)
        muse.plist.add_hook('playing', self.set_song)
        ui.add_callback('status', self.set_status)
        ui.add_callback('notice', self.notice)
        ui.add_callback('time', self.set_time)

        mid = curses.LINES/2
        self.playlist_mid = mid
        self.playlist_top = mid - (mid-2)
        self.playlist_bot = mid + (mid-2)

        if curses.LINES % 2 != 0:
            # odd number of lines, add one to bottom
            self.playlist_bot += 1

        # evil, but curses doesn't seem to do it for me?
        # os.system("(sleep 2; tput civis)&")

    def refresh(self):
        self.stdscr.refresh()

    def hook(self, fd):
        """ called when there's action on the FD we're watching - ie keypress,
        etc """

        # okay, there's action on the keyboard. read the key and parse the
        # entry.

        # if you're wondering about the weird keybindings, i use a dvorak
        # keyboard :-)

        try:
            c = self.stdscr.getch()

            if c >= 0 and c < 256:
                char = chr(c)
            else:
                char = None

            # backspace: next song in queue
            if c == 263:
                muse.player.next_with_queue()

            # tab: add current to queue
            elif c == 9:
                muse.player.add_to_queue()
                ui.notice("added to queue - %d songs waiting" % \
                          (len(playlist.queue)-1))
                muse.plist.next()

            elif char == "f":
                muse.player.fwd()
            elif char == "b":
                muse.player.back()

            elif char == " ":
                muse.player.play(muse.plist.selected)

            elif char == ';':
                muse.plist.selected = muse.plist.sel_jump(1)
            elif char == 'a':
                muse.plist.selected = muse.plist.sel_jump(-1)

            elif char == "`":
                muse.player.play_current()
                ui.notice("playing current")

            elif char == '\n':
                muse.player.play(
                    muse.plist.sel_jump_field_changed('artist', 1))

            elif char == '\\':
                muse.player.play(
                    muse.plist.sel_jump_field_changed('artist', -1))

            elif char == '-':
                muse.player.next_dir()

            elif char == '=':
                muse.player.prev_dir()

            elif char == 'o' or c == 260:
                muse.plist.prev_artist()
                ui.notice("jumping to prev artist")

            elif char == 'q' or c == 261:
                muse.plist.next_artist()
                ui.notice("jumping to next artist")

            elif char == 'e':
                muse.plist.prev_dir()
                ui.notice("prev dir: " + \
                          os.path.split(muse.plist.current()['dir'])[1])

            elif char == 'j':
                muse.plist.next_dir()
                ui.notice("next dir: " + \
                          os.path.split(muse.plist.current()['dir'])[1])

            elif c == 265:
                muse.player.start(persist=0)

            elif char == "~":
                muse.player.message("switch_player")

            elif c == 12: # C-l
                self.stdscr.clearok(1)

            elif char == 'Q':
                muse.player.quit()

            elif char == 'p':
                muse.player.pause()
        except:
            debug("attr error curses")

    def blank(self):
        """ blank - toggle the 'screen' on and off. doesn't need to be
        implemented for all interface modules"""
        pass


    def notice(self, msg):
        """ shows a low-priority message - the sort of thing to appear briefly
        in a status bar, for instance."""

        msg = msg[:(curses.COLS-2)] # strip to make sure it'll fit

        self.stdscr.move(curses.LINES-1, 0)

        self.stdscr.attron(curses.A_REVERSE)
        self.stdscr.addstr(" " * (curses.COLS-1))

        self.stdscr.addstr(curses.LINES-1, 1, msg, curses.A_REVERSE)

        self.stdscr.attroff(curses.A_REVERSE)

        # now we add a timer to clear the notice in ~ 1.5 seconds
        multiplex.set_timer("curses_notice_timer", 2, \
                            lambda x=self: x.notice(""))

    def set_time(self):
        """given a time in seconds, update the display"""

        self.stdscr.refresh()
        secs = muse.player.time_tell()

        self.stdscr.move(0, curses.COLS - 7)

        self.stdscr.attron(curses.A_REVERSE)
        self.stdscr.addstr(" " * 7)

        if secs != None:
            str = "%.2d:%.2d" % (int(secs/60),secs%60)
        else:
            str = "     "

        self.stdscr.addstr(0, curses.COLS-len(str)-1, str)

        self.stdscr.attroff(curses.A_REVERSE)

    def set_song(self):

        if muse.plist.pl != None:
            song = muse.plist.pl.song
        else:
            song = ""

        self.stdscr.move(0,0)
        self.stdscr.attron(curses.A_REVERSE)
        self.stdscr.addstr(" " * (curses.COLS - 15)) # save 15 for time, etc

        # crop string
        song = song[: curses.COLS - 14]

        # add it

        self.stdscr.addstr(0,1, song)

        self.stdscr.attroff(curses.A_REVERSE)

    def set_status(self):

        status = muse.player.status

        self.stdscr.move(0, curses.COLS - 15)
        self.stdscr.attron(curses.A_REVERSE)
        self.stdscr.addstr(" " * 9) # save 15 for time, etc

        # crop string
        status = status[:8]

        # add it

        self.stdscr.addstr(0,curses.COLS - 15, status)

        self.stdscr.attroff(curses.A_REVERSE)

    def playlist_line(self, y, song, hi=0, bold=0):
        self.stdscr.move(y, 0)

        song = song[:curses.COLS-3]

        if hi:
            self.stdscr.attron(curses.A_REVERSE)
            # pad to end of line
            song = song.ljust(curses.COLS-3)

        if bold:
            self.stdscr.attron(curses.A_BOLD)

        self.stdscr.addstr(song)

        if hi:
            self.stdscr.attroff(curses.A_REVERSE)

        if bold:
            self.stdscr.attroff(curses.A_BOLD)

#                muse.player.clear_queue()
#                muse.player.play_next()
                 #muse.player.next()
        self.stdscr.clrtoeol()

    def draw_playlist(self, queue=0):
        """ draw the normal playlist. if passed with queue=1, draw the queue
        instead """

        #debug("redraw playlist with pl_len=" + str(len(muse.plist)))

        if len(muse.plist)==0:
            self._clear_playlist()
            return

        old_playlist = None

        cur = muse.plist.selected
        top = cur - (self.playlist_mid - self.playlist_top)
        # for x in range(0, cur - self.playlist_top, cur + self.playlist_bot):
        for x in range(self.playlist_top, self.playlist_bot):
            if top >= 0 and top < len(muse.plist):

                song = muse.plist[top]['artist'] + " - "

                if muse.plist[top].has_key('track'):
                    foo = str(muse.plist[top]['track'])
                    if len(foo)==1: foo = "0" + foo
                    song = song + foo + '. '

                song += muse.plist[top]['title']

                try:
                    if muse.plist[top]['playlist'] != old_playlist and \
                       x != self.playlist_top:
                        song = ">" + song
                    else:
                        song = " " + song
                    old_playlist = muse.plist[top]['playlist']

                except KeyError:
                    song = " " + song



                if x == self.playlist_mid:
                    self.playlist_line(x, song, 1)
                else:
                    pass
                    self.playlist_line(x, song)

                if x == (self.playlist_mid-1):
                    pass
                    #self.playlist_line(x, song, bold=1)
                else:
                    self.playlist_line(x, song)
            else:
                self.playlist_line(x, "")

            top += 1


        self.refresh()


    def _clear_playlist(self):
        for x in range(self.playlist_top, self.playlist_bot):
            self.stdscr.move(x, 0)
            self.stdscr.clrtoeol()














