#!/usr/bin/env python
#
# $Id: ui_orbital.py,v 1.1.1.1 2003/02/11 06:07:41 resolve Exp $
#
# Started 5/8/01 by Damien Elmes <resolve@repose.cx>
# This file is licensed under the GPL.
#

"""
Plugin for muse which delivers song information on a Matrix Orbital LCD screen,
and keypad watching, for remote commands.
"""

import os
import types

import ui
import multiplex
import muse
import config
import orbital
import control

from debug import debug

PLUGIN_TYPE="ui"

config_defaults = {
    "lcd_device":    "/dev/ttyS0",
    "baud":          "19200",
    "screen_width":  "40",
    "screen_height": "2"
    }

def display_menu():
    return

keymap = {
    "R": 1,      # key layout:
    "Q": 2,      #
    "P": 3,      #    1 2 3
    "M": 4,      #    4 5 6
    "L": 5,      #    7 8 9
    "K": 6,      #    * 0 #
    "H": 7,      #
    "G": 8,      #    3 6 9 #
    "F": 9,      #    2 5 8 0
    "B": 0,      #    1 4 7 *
    "C": '*',    #
    "A": '#'     #
    }

# put back prev-d

class new(ui.plugin):

    def __init__(self):

        ui.plugin.__init__(self)

        self.plugin_type = "ui"
        self.plugin_desc = "matrix orbital LCD support"
        self.plugin_name = "orbital"

        config.set_defaults("ui_orbital", config_defaults)

        self.lcd_device = config.get("ui_orbital", "lcd_device")
        baud = config.get("ui_orbital", "baud")

        try:
            self.orb = orbital.Orbital(device=self.lcd_device, baud=baud)
        except orbital.error:
            debug("Couldn't open LCD device, using emulation.")
            self.orb = orbital.OrbitalTester()

        self.old_orb = None

        muse.plist.add_hook('selected', self.draw_playlist)

        self.screen_w = int(config.get("ui_orbital", "screen_width"))
        self.screen_h = int(config.get("ui_orbital", "screen_height"))

        muse.plist.add_hook('playing', self.set_song)
        ui.add_callback('status', self.set_status)
        ui.add_callback('time', self.set_time)
        ui.add_callback('notice', self.notice)
        control.add_cb(self.on_control)

        self.orb.set_keypad_cb(self.on_keypad)

        # for play, pause, etc
        self.setup_custom_chars()

        self.status_lock = 0

        self.orb.clear_screen()
        self.orb.write((" " * 40) + "muse v0.1.9 loading.")

        # cmd section
        out = " " * 5
        self.orb.write(out, self.screen_w-5, 2, 5, 1)

        # used for storing information about which menu the user is on, etc
        self.state = {}
        self.state['mode']='song'
        self.state['song_jump'] = "immediate"

        self.song_keymap = {
            1:   lambda self=self: self.change_song("prev"),
            2:   lambda self=self: self.change_song("next art"),
            3:   lambda self=self: self.change_song("sel"),
            4:   lambda self=self: self.change_song("prev dir"),
            5:   lambda self=self: self.change_song("prev art"),
            6:   lambda self=self: self.change_song("next dir"),
            7:   lambda self=self: None,
            8:   display_menu,
            9:   lambda self=self: control.cmd("pause"),
            0:   lambda self=self: self.change_song("fwd"),
            '*': lambda self=self: self.change_song("back"),
            '#': "ok"
            }



    def setup_custom_chars(self):
        """ register some custom chars used for muse """

        play  =  ( 0, 8, 12, 14, 14, 12, 8, 0 )
        pause =  ( 0, 9, 9, 9, 9, 9, 9, 0 )
        stop  =  ( 0, 0, 14, 14, 14, 14, 0, 0 )
        fwd   =  ( 0, 4, 6, 31, 31, 6, 4, 0 )
        back   =  ( 0, 4, 12, 31, 31, 12, 4, 0 )

        self.orb.set_custom_char(0, play)
        self.orb.set_custom_char(1, pause)
        self.orb.set_custom_char(2, stop)
        self.orb.set_custom_char(3, fwd)
        self.orb.set_custom_char(4, back)

        self.charmap = { "play"   : chr(0),
                         "pause"  : chr(1),
                         "stop"   : chr(2),
                         "fwd"    : chr(3),
                         "back"   : chr(4)  }

    def on_keypad(self, c):
        """ called when there's action on the FD we're watching - ie keypress,
        etc """

        # okay, there's action on the keypad. read the key and parse the
        # entry.

        if self.state['mode']=='song':
            self.songmode(keymap[c])

    def on_control(self, cmd):
        """ called when a change is made by the user """

        if cmd.find("next")==0:
            self.orb.write(self.charmap['fwd'], self.screen_w, 2)
        elif cmd.find("prev")==0:
            self.orb.write(self.charmap['back'], self.screen_w, 2)

        if cmd == "next" or cmd == "prev" or \
               cmd == "nextp" or cmd == "prevp":
            out = "song"
        elif cmd[-2:] == "-a":
            out = "art"
        elif cmd[-2:] == "-d":
            out = "dir"
        else:
            out = cmd[:4]


        out = " " + out
        self.orb.write(out, self.screen_w-5, 2, 5, 1)

        self.status_lock = 1
        # at the end, timer to refresh section
        multiplex.set_timer("orbital_control_timer", 2, \
                            self.clear_control)

    def clear_control(self):
        """ clear the control area and update status """

        self.status_lock = 0
        self.orb.write("", self.screen_w-4, 2, 5, 1)
        ui.new_status()

    def notice(self, msg):
        """ shows a low-priority message - the sort of thing to appear briefly
        in a status bar, for instance."""

        self.orb.write(msg, 1, 2, self.screen_w-6, 1)

        # now we add a timer to clear the notice in ~ 1.5 seconds
        multiplex.set_timer("orbital_notice_timer", 1.5, \
                            self.set_song)

    def set_time(self):
        """given a time in seconds, update the display"""

        secs = muse.player.time_tell()

        if secs != None:
            str = " %.2d:%.2d" % (int(secs/60),secs%60)
        else:
            str = " 00:00"

        self.orb.write(str, self.screen_w-5, 1, 6)

    def set_status(self):
        if self.status_lock: return

        status = muse.player.status

        if status == "playing":
            self.orb.write(self.charmap['play'], self.screen_w, 2)

        if status == "paused":
            self.orb.write(self.charmap['pause'], self.screen_w, 2)

        #if status == "stopped":
        #    self.orb.write(self.charmap['stop'], self.screen_w, 2)

    def set_song(self):
        if not muse.player.playing():
            self.orb.write("", 1, 1, self.screen_w-6, pad=1)
            self.orb.write("", 1, 2, self.screen_w-6, pad=1)
            return

        # evil hack, fixme
        artist = muse.plist.playing['artist']
        title = muse.plist.playing['title']

        self.orb.write(artist, 1, 1, self.screen_w-6, pad=1)
        self.orb.write(title, 1, 2, self.screen_w-6, pad=1)

    def playlist_line(self, y, song, hi=0):
        self.stdscr.move(y, 1)

        song = song[:curses.COLS-3]

        if hi:
            self.stdscr.attron(curses.A_REVERSE)
            # pad to end of line
            song = song.ljust(curses.COLS-3)

        self.stdscr.addstr(song)

        if hi:
            self.stdscr.attroff(curses.A_REVERSE)

#                muse.player.clear_queue()
#                muse.player.play_next()
                 #muse.player.next()
        self.stdscr.clrtoeol()

    def songmode(self, code):
        """ received a key while in song mode """

        ent = self.song_keymap[code]

        if isinstance(ent, types.FunctionType):
            ent()
        elif isinstance(ent, types.StringType):
            if ent == "ok":
                # if we're in immediate mode, toggle to delayed
                if self.state['song_jump']=="immediate":
                    self.notice("switching to delayed mode")
                    self.state['song_jump']="delayed"
                    return

                # if we're in delayed mode
                if self.state['song_jump']=="delayed":

                    # and they've changed the song
                    if muse.plist.dirty:
                        # then jump to it
                        control.cmd("cur")
                    else:
                        # or if they haven't, switch back to immediate
                        self.state['song_jump']="immediate"
                        self.notice("switching to immediate mode")

    def change_song(self, cmd):
        """ switch to a new song, or just change the queue, depending on the
        mode """

        if self.state['song_jump'] == "immediate" or \
               (cmd[0] != 'p' and cmd[0] != 'n'):
            control.cmd(cmd)
        elif self.state['song_jump'] == "delayed":
            # append p to change playlist, not song
            cmd = cmd[0:4] + 'p' + cmd[4:]
            control.cmd(cmd)
            self.set_song()


    def draw_playlist(self, queue=0):
        pass

    def draw_artist(self):
        """ draw a list of the artist available to jump to """
        pass

    def _clear_playlist(self):
        pass












